import unittest
from nasti.hooks import Hooks
import nasti.exceptions as exceptions
from nasti.nastifile import NastiFile
import os

class TestHooks(unittest.TestCase):

    def test_null_hooks(self):
        hooks = Hooks({})
        self.assertEqual(hooks.before, False)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, False)
        self.assertEqual(hooks.run_before(), False)
        self.assertEqual(hooks.run_before(), False)
    
    def test_hooks_before_with_no_cleanup(self):
        script = "tests/hooks/success.sh"
        hooks = Hooks({
            "before_script": script,
        })
        self.assertEqual(hooks.before, script)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, False)
        self.assertEqual(hooks.run_before(), True)
    
    def test_hooks_after_with_no_cleanup(self):
        script = "tests/hooks/success.sh"
        hooks = Hooks({
            "after_script": script,
        })
        self.assertEqual(hooks.before, False)
        self.assertEqual(hooks.after, script)
        self.assertEqual(hooks.auto_cleanup, False)
        self.assertEqual(hooks.run_after(), True)
    
    def test_hooks_before_nonzero_exit(self):
        script = "tests/hooks/fail.sh"
        hooks = Hooks({
            "before_script": script,
        })
        self.assertEqual(hooks.before, script)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, False)
        with self.assertRaises(exceptions.HooksScriptExecutionFailed):
            hooks.run_before()
    
    def test_hooks_after_nonzero_exit(self):
        script = "tests/hooks/fail.sh"
        hooks = Hooks({
            "after_script": script,
        })
        self.assertEqual(hooks.before, False)
        self.assertEqual(hooks.after, script)
        self.assertEqual(hooks.auto_cleanup, False)
        with self.assertRaises(exceptions.HooksScriptExecutionFailed):
            hooks.run_after()

    def test_hooks_before_script_doesnt_exist(self):
        script = "tests/hooks/doesntexist.sh"
        hooks = Hooks({
            "before_script": script,
        })
        self.assertEqual(hooks.before, script)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, False)
        with self.assertRaises(exceptions.HooksScriptNotFound):
            hooks.run_before()
    
    def test_hooks_after_script_doesnt_exist(self):
        script = "tests/hooks/doesntexist.sh"
        hooks = Hooks({
            "after_script": script,
        })
        self.assertEqual(hooks.before, False)
        self.assertEqual(hooks.after, script)
        self.assertEqual(hooks.auto_cleanup, False)
        with self.assertRaises(exceptions.HooksScriptNotFound):
            hooks.run_after()
    
    def test_hooks_before_script_with_cleanup(self):
        # create a copy of the success script
        os.system("cp tests/hooks/success.sh tests/hooks/success_copy.sh")

        script = "tests/hooks/success_copy.sh"
        hooks = Hooks({
            "before_script": script,
            "auto_cleanup": True
        })
        self.assertEqual(hooks.before, script)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, True)
        self.assertEqual(hooks.run_before(), True)

        # verify the script was cleaned up
        self.assertEqual(os.path.exists(script), False)
    
    def test_hooks_after_script_with_cleanup(self):
        # create a copy of the success script
        os.system("cp tests/hooks/success.sh tests/hooks/success_copy.sh")

        script = "tests/hooks/success_copy.sh"
        hooks = Hooks({
            "after_script": script,
            "auto_cleanup": True
        })
        self.assertEqual(hooks.before, False)
        self.assertEqual(hooks.after, script)
        self.assertEqual(hooks.auto_cleanup, True)
        self.assertEqual(hooks.run_after(), True)

        # verify the script was cleaned up
        self.assertEqual(os.path.exists(script), False)
    
    def test_hooks_before_script_with_cleanup_failure(self):
        script = "popo.sh"
        hooks = Hooks({
            "before_script": script,
            "auto_cleanup": True
        })
        self.assertEqual(hooks.before, script)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, True)

        with self.assertRaises(exceptions.HooksCleanupFailed):
            hooks.cleanup(script)

    def test_hooks_script_execute_os_error(self):
        script = "tests/hooks/no_executable"
        hooks = Hooks({
            "before_script": script,
        })
        self.assertEqual(hooks.before, script)
        self.assertEqual(hooks.after, False)
        self.assertEqual(hooks.auto_cleanup, False)
        with self.assertRaises(exceptions.HooksScriptExecutionFailed):
            hooks.run_before()
    
    def test_hooks_end_to_end(self):
        # Create a copy of the success script called before
        os.system("cp tests/hooks/success.sh tests/hooks/before_script.sh")
        # Create a copy of the success script called after
        os.system("cp tests/hooks/success.sh tests/hooks/after_script.sh")

        input_dep = func = lambda x: "input-from-user"
        print_dep = func = lambda x: None
        nasti_file = NastiFile({
            "path": "tests/hooks",
            "os_dep": os,
            "open_dep": open,
            "input_dep": input_dep,
            "print_dep": print_dep
        })
        nasti_file.load()
        nasti_file.run()

        # verify the before script was cleaned up
        self.assertEqual(os.path.exists("tests/hooks/before_script.sh"), False)
        # verify the after script was cleaned up
        self.assertEqual(os.path.exists("tests/hooks/after_script.sh"), False)


import nasti.exceptions as exceptions
import os

class Hooks:
    BEFORE_KEY = "before_script"
    AFTER_KEY = "after_script"
    AUTO_CLEANUP_KEY = "auto_cleanup"
    OS_DEP_KEY = "os_dep"
    WORKING_DIR_KEY = "working_dir"

    before = False
    after = False
    auto_cleanup = False
    os_dep = os
    cleanup_command = "rm "

    def __init__(self, opts):
        self.working_dir = os.getcwd()
        if self.BEFORE_KEY in opts:
            self.before = opts[self.BEFORE_KEY]
        if self.AFTER_KEY in opts:
            self.after = opts[self.AFTER_KEY]
        if self.AUTO_CLEANUP_KEY in opts:
            self.auto_cleanup = opts[self.AUTO_CLEANUP_KEY]
        if self.OS_DEP_KEY in opts:
            self.os_dep = opts[self.OS_DEP_KEY]
        if self.WORKING_DIR_KEY in opts:
            self.working_dir = opts[self.WORKING_DIR_KEY]

    def __run_script(self, script):
        # Verify the script exists
        if not self.os_dep.path.exists(script):
            raise exceptions.HooksScriptNotFound(f"Hooks script not found: {script}")
        try:
            result = self.os_dep.system(f"sh {script} > /dev/null 2>&1")
        except Exception as e:
            raise exceptions.HooksScriptExecutionFailed(f"Hooks script failed: {e}")
        if result != 0:
            raise exceptions.HooksScriptExecutionFailed("Hooks script failed.")

    def cleanup(self, script):
        if self.auto_cleanup == False:
            return
        try:
            result = self.os_dep.system(f"{self.cleanup_command} {script} > /dev/null 2>&1")
        except Exception as e:
            raise exceptions.HooksCleanupFailed(f"Hooks cleanup failed: {e}")
        if result != 0:
            raise exceptions.HooksCleanupFailed("Hooks cleanup failed.")

    def __run_hook(self, hook):
        if hook == False:
            return hook
        self.origina_cwd = self.os_dep.getcwd()
        self.os_dep.chdir(self.working_dir)
        self.__run_script(hook)
        self.cleanup(hook)
        self.os_dep.chdir(self.origina_cwd)
        return True

    def run_before(self):
        return self.__run_hook(self.before)

    def run_after(self):
        return self.__run_hook(self.after)


import os
import sys

# Took this idea from https://stackoverflow.com/questions/1896918/running-unittest-with-typical-test-directory-structure
# I couldn't figure out how to get a sane import path for the tests without this.
PROJECT_PATH = os.getcwd()
SOURCE_PATH = os.path.join(
    PROJECT_PATH,"nasti"
)
sys.path.append(SOURCE_PATH)


import subprocess

from nasti.source_handlers import SourceHandlerResolver
from nasti.nastifile import NastiFile

class Nasti:
    
    def __init__(self, opts={}):
        # Dependency injection
        self.os_dep = opts["os_dep"]
        self.open_dep = opts["open_dep"]
        self.input_dep = opts["input_dep"]
        self.help_text = opts["help_text"]

        self.source = opts["source"]
        self.handler = None
        self.print_dep = opts["print_dep"]
        self.output_dir = opts["output_dir"]
        self.git_init = opts["git_init"]
        self.accept_defaults = False

        self.silent_opts = {}
        self.silent_mode = False
        if "silent_opts" in opts:
            self.silent_opts = opts["silent_opts"]
        if "silent_mode" in opts:
            self.silent_mode = opts["silent_mode"]
        if "accept_defaults" in opts:
            self.accept_defaults = opts["accept_defaults"]

    def run(self):
        try:
            self.__get_source()
            self.__create_output_dir()
            self.__copy_source_files()
            self.__load_nasti_file()
            self.nasti_file.validate_mutations()
            self.nasti_file.run()
            self.__clean_up()
            self.__git_init()
        except Exception as e:
            self.handler.clean_up()
            self.delete_output_dir()
            # Print a pretty error
            self.print_dep(f"An error ocurred processing the template: ")
            raise e

    def __clean_up(self):
        #delete the nastifile
        self.os_dep.system(f"rm {self.output_dir}/nasti.yaml")
        # delete the git repo
        self.os_dep.system(f"rm -rf {self.output_dir}/.git")

    def __git_init(self):
        if self.git_init:
            with self.open_dep(self.os_dep.devnull, 'w') as devnull:
                return subprocess.run(f"cd {self.output_dir} && git init && git add -A && git commit -am 'Initial commit'", stdout=devnull, stderr=devnull, shell=True)


    def __create_output_dir(self):
        attemps = 0
        max_attempts = 3
        while True:
            if not self.output_dir:
                self.output_dir = input("Enter an output directory name: ")
            try:
                # Attempt to create the directory
                self.os_dep.makedirs(self.output_dir)
                if not self.silent_mode:
                    self.print_dep(f"Directory '{self.output_dir}' created successfully.")
                break
            except OSError as e:
                # Handle error if directory creation fails
                self.print_dep(f"Error: {e}")
                self.print_dep("Please check the directory name and try again.")
                self.output_dir = None
                attemps += 1
                if attemps >= max_attempts:
                    raise Exception("Error: Something really weird is up. ")

    def delete_output_dir(self):
        self.os_dep.system(f"rm -rf {self.output_dir}")

    def __copy_source_files(self):
        # Copy the files from the source to the output dir
        self.os_dep.system(f"cp -r {self.handler.source_dir}/* {self.output_dir}")

    def __load_nasti_file(self):
        if not self.handler:
            raise Exception("Error: No source handler found.")
        self.nasti_file = NastiFile({
            "path": self.output_dir,
            "os_dep": self.os_dep,
            "open_dep": self.open_dep,
            "print_dep": self.print_dep,
            "input_dep": self.input_dep,
            "accept_defaults": self.accept_defaults,
            "silent_mode": self.silent_mode,
            "silent_opts": self.silent_opts,
        })
        self.nasti_file.load()

    def __get_source(self):
        resolver = SourceHandlerResolver(self.source, self.help_text, self.os_dep, self.print_dep)
        self.handler = resolver.resolve()
        self.handler.run()




import unittest
from nasti.globals import Global
import nasti.exceptions as exceptions

class TestGlobal(unittest.TestCase):
    def test_global_with_full_config(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "help": "test help",
            "validation": {
                "regex": ".*"
            }
        }
        global_obj = Global(global_config, input_dep=input, print_dep=print)
        self.assertEqual(global_obj.name, "test")

    def test_global_with_required_config(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
        }
        global_obj = Global(global_config, input_dep=input, print_dep=print)
        self.assertEqual(global_obj.name, "test")
    
    def test_global_with_missing_required_config(self):
        global_config = {
            "prompt": "test prompt",
        }
        with self.assertRaises(exceptions.GlobalRequiredKeysMissingException):
            Global(global_config, input_dep=input, print_dep=print)

    def test_global_with_misconfigured_validation(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "validation": {
                "regex": ".*",
                "kind": "bogus"
            }
        }
        # Validations raise multiple exceptions
        # We only test for an exception here
        with self.assertRaises(Exception):
            Global(global_config, input_dep=input, print_dep=print)

    def test_global_with_too_many_tries(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "validation": {
                "kind": "url"
            }
        }
        input_dep = func = lambda x: "bogus input"
        print_dep = func = lambda x: None
        global_obj = Global(global_config, input_dep=input_dep, print_dep=print_dep)
        with self.assertRaises(exceptions.GlobalTooManyInputTriesException):
            global_obj.populate()
    
    def test_global_with_validation(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "validation": {
                "kind": "url"
            }
        }
        input_dep = func = lambda x: "http://www.google.com"
        print_dep = func = lambda x: None
        global_obj = Global(global_config, input_dep=input_dep, print_dep=print_dep)
        global_obj.populate()
        self.assertTrue(global_obj.validate())

    def test_global_with_validation_silent(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "validation": {
                "kind": "url"
            }
        }
        input_dep = func = lambda x: None
        print_dep = func = lambda x: None
        global_obj = Global(global_config, input_dep=input_dep, print_dep=print_dep, silent_mode=True, silent_opts={"test": "http://www.google.com"})
        global_obj.populate()
        self.assertTrue(global_obj.validate())

    def test_global_with_validation_failure(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "validation": {
                "kind": "url"
            }
        }
        input_dep = func = lambda x: "https://www.google.com"
        print_dep = func = lambda x: None
        global_obj = Global(global_config, input_dep=input_dep, print_dep=print_dep)
        global_obj.populate()
        global_obj.value = "bogus"
        self.assertFalse(global_obj.validate())

    def test_global_with_validation_failure_silent(self):
        global_config = {
            "name": "test",
            "prompt": "test prompt",
            "validation": {
                "kind": "url"
            }
        }
        input_dep = func = lambda x: None
        print_dep = func = lambda x: None
        global_obj = Global(global_config, input_dep=input_dep, print_dep=print_dep, silent_mode=True, silent_opts={"test": "coheed and cambria"})
        with self.assertRaises(exceptions.GlobalSilentModeException):
            global_obj.populate()


import os
import re
import yaml

from nasti.mutation import Mutation
from nasti.globals import Global
import nasti.exceptions as exceptions
from nasti.hooks import Hooks

# This class is used to store the results of the find command
class UnmentionedFilesResult:
    def __init__(self):
        self.results = []
    
    def add(self, mutation, files):
        if len(files) == 0:
            return
        self.results.append(UnmentionedFilesResultItem(files, mutation))

    def get_report(self):
        if len(self.results) == 0:
            return ""
        report = ""
        report += "\n:exclamation_mark: The following mutations [green]match[/green] files [red]not listed[/red] in the nastifile:\n"
        for result in self.results:
            report += result.get_report()
        return report

    def get_results(self):
        return self.results

# This class stores the individual results of a find operation
class UnmentionedFilesResultItem:
    def __init__(self, files, mutation):
        self.files = files
        self.mutation = mutation

    def get_report(self):
        report = ""
        report += f"\nMutation [magenta]{self.mutation.name}[/magenta] [green]matches[/green] but [red]does not[/red] reference:"
        for file in self.files:
            report += f"\n    • [blue]{file}[/blue]"
        return report
    
    def get_files(self):
        return self.files

    def get_mutation(self):
        return self.mutation

class NastiFile:
    MUTATIONS_KEY="mutations"
    NAME_KEY="name"
    DEFAULT_KEY="default"
    PROMPT_KEY="prompt"
    REPLACE_KEY="replace"
    FILES_KEY="files"
    HELP_KEY="help"
    VALIDATION_KEY="validation"
    GLOBALS_KEY="globals"
    GREETING_KEY="greeting"
    HOOKS_KEY="hooks"

    # This is a dictionary of global variables
    # This is set by running the globals
    globals = {}

    def __init__(self, opts={}):
        # Dependency injection
        self.os_dep = opts["os_dep"]
        self.open_dep = opts["open_dep"]
        self.accept_defaults = False
        if "accept_defaults" in opts:
            self.accept_defaults = opts["accept_defaults"]
        # Null object pattern
        self.hooks = Hooks({})
        if "print_dep" in opts:
            self.print_dep = opts["print_dep"]
        else:
            self.print_dep = print
        if "input_dep" in opts:
            self.input_dep = opts["input_dep"]
        else:
            self.input_dep = input
        self.silent_opts = {}
        self.silent_mode = False
        if "silent_opts" in opts:
            self.silent_opts = opts["silent_opts"]
        if "silent_mode" in opts:
            self.silent_mode = opts["silent_mode"]
        self.working_dir = opts["path"]
        self.__set_path(opts["path"])
        

    def load(self):
        self.__verify_exists()
        self.init_config()
        self.init_hooks()

    def run(self):
        self.load()
        self.run_greeting()
        self.hooks.run_before()
        self.run_globals()
        self.run_mutations()
        self.hooks.run_after()

    def init_hooks(self):
        if self.HOOKS_KEY in self.config:
            hooks_opts = self.config[self.HOOKS_KEY]
            hooks_opts["working_dir"] = self.working_dir
            self.hooks = Hooks(self.config[self.HOOKS_KEY])

    def init_config(self):
        with self.open_dep(self.path, 'r') as file:
            try:
                self.config = yaml.load(file, Loader=yaml.SafeLoader)
            except yaml.YAMLError as e:
                raise exceptions.NastiFileInvalidYamlException(f"Error: Unable to load {self.path}.")

    def run_greeting(self):
        if self.GREETING_KEY in self.config and not self.silent_mode:
            self.print_dep(self.config[self.GREETING_KEY])

    def run_mutations(self):
        working_dir = self.get_dir()
        for mutation_config in self.config[self.MUTATIONS_KEY]:
            mutation_config["globals"] = self.globals
            mutation = Mutation(mutation_config, working_dir, os, open, self.input_dep, self.print_dep, self.accept_defaults, self.silent_mode, self.silent_opts)
            mutation.run()

    def run_globals(self):
        if self.GLOBALS_KEY in self.config:
            for global_config in self.config[self.GLOBALS_KEY]:
                global_obj = Global(global_config, self.input_dep, self.print_dep, self.silent_mode, self.silent_opts)
                global_obj.populate()
                self.globals[global_obj.get_name()] = global_obj.get_value()

    def get_global(self, name):
        if name in self.globals:
            return self.globals[name]
        raise exceptions.NastiFileGlobalNotFoundException(f"Error: Global {name} not found.")

    # Get the abolute path of the directory containing the nasti file
    def get_dir(self):
        return self.os_dep.path.dirname(self.os_dep.path.abspath(self.path))
    
    # Validate the mutations in the nastifile
    def validate_mutations(self):
        for mutation in self.__mutations():
             mutation.validate()
    
    # Find files that are not mentioned in the nastifile
    def find_unmentioned_files(self):
        unmentioned_files = UnmentionedFilesResult()
        for mutation in self.__mutations():
            mutation_unmentioned_files = mutation.find_unmentioned_files(self.get_dir())
            unmentioned_files.add(mutation, mutation_unmentioned_files)
        return unmentioned_files

    # Get a list of mutation objects from the config
    # Does some validation on the config
    def __mutations(self):
        mutations = []
        self.load()
        # verify there are mutations
        if not self.MUTATIONS_KEY in self.config:
            raise exceptions.NastiFileNoMutationsException(f"Error: {self.path} does not contain any mutations.")
        working_dir = self.get_dir()
        for mutation_config in self.config[self.MUTATIONS_KEY]:
            # verify each mutation is valid
            # throws an exception if there is a problem
            self.__validate_mutation_config_keys(mutation_config)
            # create a mutation object and add it to the list
            mutation = Mutation(mutation_config, working_dir)
            mutations.append(mutation)
        return mutations

    def __validate_mutation_config_keys(self, mutation_config):
        valid_keys = [
            self.NAME_KEY, 
            self.PROMPT_KEY, 
            self.REPLACE_KEY, 
            self.FILES_KEY, 
            self.HELP_KEY, 
            self.VALIDATION_KEY,
            self.DEFAULT_KEY,
        ]
        # Exit the app
        for key in mutation_config:
            if not key in valid_keys:
                raise exceptions.NastiFileUnknownKeysException(f"Error: invalid key in mutation config: {key}")

    def __set_path(self, path):
        if not path:
            path = './nasti.yaml'
        self.path = f"{path}/nasti.yaml"


    def __verify_exists(self):
        # ensure that the file exists
        if not self.os_dep.path.isfile(self.path):
            raise exceptions.NastiFileUnableToOpenFileException(f"Error: {self.path} does not exist.")



import click
import sys
import os
import json
import yaml
from nasti.nastifile import NastiFile
from nasti.nasti import Nasti
import rich
from prompt_toolkit import PromptSession
from prompt_toolkit.history import InMemoryHistory

def remove_silent_opt_outer_quotes(s):
    if len(s) < 2:  # If the string length is less than 2, just return the original string
        return s
    if (s[0] == "'" and s[-1] == "'") or (s[0] == '"' and s[-1] == '"'):
        return s[1:-1]
    return s

def parse_silent_opts(silent_str):
    silent_dict = {}
    pairs = silent_str.split(',')

    for pair in pairs:
        if '=' not in pair:
            raise ValueError(f"{pair}. Expected format: key=value.")

        key, value = pair.split('=')

        if not key or not value:
            raise ValueError(f"{pair}. Neither key nor value can be empty.")

        silent_dict[key] = remove_silent_opt_outer_quotes(value)

    return silent_dict

@click.group()
def cli():
    """
    Welcome to NASTI!

    NASTI is free and open source software. Feel free to get weird with it!
    """
    pass

@click.command()
@click.argument("source", required=True)
@click.argument("dest_dir", required=False)
@click.option("--git", "-g", help="Create a git repo in the new project. Default is True.", is_flag=True, default=True )
@click.option("--defaults", "-d", help="Accept all defaults. Default is False", is_flag=True, default=False )
@click.option("--silent", "-s", help="Silent mode with key=value pairs separated by commas.")
@click.option("--input-file", "-f", type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True), help="Input file in JSON or YAML format containing key=value pairs.")
def process(source, git, defaults, dest_dir, silent, input_file):
    # When in silent mode we don't prompt the user for input
    silent_mode = False
    silent_opts = {}
    if silent:
        try:
            silent_opts = parse_silent_opts(silent)
            if len(silent_opts) == 0:
                raise ValueError("Silent mode key/value pairs cannot be empty.")
            silent_mode = True
        except ValueError as e:
            rich.print("[red]:stop_sign:[bold] Error processing silent mode key/value pairs[/bold][red]")
            rich.print(e)
            return

    if input_file:
        _, ext = os.path.splitext(input_file)
        try:
            with open(input_file, 'r') as f:
                if ext in ['.yaml', '.yml']:
                    silent_opts = yaml.safe_load(f)
                elif ext == '.json':
                    silent_opts = json.load(f)
                else:
                    print(f"[red]:stop_sign:[bold] Unsupported file format {ext}. Only .json and .yaml are supported.[/bold][red]")
                    return
        except yaml.YAMLError:
            print("[red]:stop_sign:[bold] Error parsing the YAML file.[/bold][red]")
            return
        except json.JSONDecodeError:
            print("[red]:stop_sign:[bold] Error parsing the JSON file.[/bold][red]")
            return
        except Exception as e:
            print(f"[red]:stop_sign:[bold] Error reading the file: {e}[/bold][red]")
            return

        if not silent_opts or not all(isinstance(key, str) and isinstance(val, str) for key, val in silent_opts.items()):
            print("[red]:stop_sign:[bold] Error in file content. Expected simple key=value pairs of strings.[/bold][red]")
            return

        silent_mode = True


    try:
        history = InMemoryHistory()
        session = PromptSession(history=history)
        nasti = Nasti({
            "source": source, 
            "print_dep": rich.print, 
            "git_init": git, 
            "help_text": cli.get_help(click.Context(cli)),
            "os_dep": os,
            "open_dep": open,
            "input_dep": session.prompt,
            "accept_defaults": defaults,
            "output_dir": dest_dir,
            "silent_mode": silent_mode,
            "silent_opts": silent_opts,
        })
        nasti.run()
    except Exception as e:
        print(f"    Error: {e}")
    except KeyboardInterrupt:
        # Check if the output directory exists
        if nasti.output_dir:
            rich.print("")
            rich.print("[red]:stop_sign:[bold] Keyboard interrupt detected.[/bold][red]")
            rich.print("[gray][italic]   Deleting output directory...[/italic][/gray]")
            nasti.delete_output_dir()
            rich.print("[gray][italic]   Exiting.[gray][italic]")

@click.command()
@click.argument("path", required=False)
def validate(path):
    if not path:
        path = "."
    try:
        nasti_file = NastiFile({
            "print_dep": rich.print,
            "path": path,
            "os_dep": os,
            "open_dep": open,
        })
        nasti_file.load()
        nasti_file.validate_mutations()
        rich.print("[green]:heavy_check_mark: Nastifile is valid.[/green]")
    except Exception as e:
        rich.print("[red]:x: Nastifile is invalid.[/red]")
        rich.print(e)

@click.command()
@click.argument("path", required=False)
def find(path):
    rich.print("[blue]Searching for files that match mutations but aren't mentioned in Nastifile...[/blue]")
    if not path:
        path = "."
    try:
        nasti_file = NastiFile({
            "print_dep": rich.print,
            "path": path,
            "os_dep": os,
            "open_dep": open,
        })
        unmentioned_files = nasti_file.find_unmentioned_files()
        rich.print(unmentioned_files.get_report())
    except Exception as e:
        rich.print(e)

cli.add_command(process)
cli.add_command(validate)
cli.add_command(find)


from nasti.validation import Validation
import unittest
import nasti.exceptions as exceptions

class TestValidation(unittest.TestCase):
    def test_valid_regex(self):
        validation = Validation({
            "regex": r'^[A-Za-z\s]+$',
        })
        assert validation.validate("Welcome To The Machine") == True

    def test_invalid_regex(self):
        validation = Validation({
            "regex": r'^[A-Za-z\s]+$',
        })
        assert validation.validate("867-5309") == False
    
    def test_missing_config(self):
        with self.assertRaises(exceptions.ValidationConfigMissingException):
            validation = Validation({
                "invalid": "invalid",
            })

    def test_invalid_config(self):
        with self.assertRaises(exceptions.ValidationConfigInvalidException):
            validation = Validation({
                "regex": r'^[A-Za-z\s]+$',
                "kind": "slug"
            })

    def test_valid_kind(self):
        validation = Validation({
            "kind": "slug",
        })
        assert validation.validate("coheed-and-cambria") == True
    
    def test_invalid_kind(self):
        validation = Validation({
            "kind": "slug",
        })
        assert validation.validate("Tonight, the south is on fire!") == False
    
    def test_unknown_kind(self):
        with self.assertRaises(exceptions.ValidationUnknownKindException):
            validation = Validation({
                "kind": "ire_works",
            })


import unittest
from nasti.mutation import Mutation
import nasti.exceptions as exceptions
import tests.mocks as mocks
import os
import yaml

class TestMutation(unittest.TestCase):
    # This tests that the files array is present but empty
    def test_empty_files(self):
        #open the yaml file
        with open("tests/nastifiles/mutation_empty_files/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_empty_files")
        with self.assertRaises(exceptions.MutationEmptyFilesException):
            mutation.validate()

    def test_unmentioned_files(self):
        #open the yaml file
        with open("tests/nastifiles/mutation_unmentioned_files/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_unmentioned_files")
        unmentioned_files = mutation.find_unmentioned_files("tests/nastifiles/mutation_unmentioned_files")
        assert set(unmentioned_files) == set(['files/nested/unmentioned', 'files/unmentioned'])

    # this tests that the files array is not present
    def test_no_files(self):
        #open the yaml file
        with open("tests/nastifiles/mutation_no_files/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        with self.assertRaises(exceptions.MutationRequiredKeysMissingException):
            Mutation(mutation_config, "tests/nastifiles/mutation_no_files/")

    # this tests that the files array is present but the files are not on disk
    def test_files_dont_exist(self):
        #open the yaml file
        with open("tests/nastifiles/mutation_files_dont_exist/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_files_dont_exist", os, open)
        with self.assertRaises(exceptions.MutationFileDoesNotExistException):
            mutation.validate()
    
    def test_too_many_tries(self):
        input_dep = func = lambda x: "bogus input"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/valid/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_too_many_tries", os, open, input_dep, print_dep)
        with self.assertRaises(exceptions.MutationTooManyInputTriesException):
            mutation.run()
    
    def test_misconfigured_validation(self):
        #open the yaml file
        with open("tests/nastifiles/mutation_misconfigured_validation/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        # Validations raise multiple exceptions
        # We only test for an exception here
        # The validation tests cover all of the various kinds of failures
        with self.assertRaises(Exception):
            Mutation(mutation_config, "tests/nastifiles/mutation_misconfigured_validation")

    def test_mutation_with_missing_required_config(self):
        mutation_config = {
            "prompt": "test prompt",
        }
        with self.assertRaises(exceptions.MutationRequiredKeysMissingException):
            Mutation(mutation_config, "")

    def test_file_doesnt_contain_replacement_text(self):
        input_dep = func = lambda x: "bogus_slug"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/mutation_file_doesnt_contain_replacement_text/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_file_doesnt_contain_replacement_text", os, open, input_dep, print_dep)
        with self.assertRaises(exceptions.MutationFileDoesNotContainReplacementStringException):
            mutation.validate()
    
    def test_text_replacement_fails(self):
        input_dep = func = lambda x: "bogus_slug"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/mutation_text_replacement_fails/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_text_replacement_fails", os, mocks.mock_failed_open, input_dep, print_dep)
        with self.assertRaises(exceptions.MutationTooManyInputTriesException):
            mutation.run()

    def test_run(self):
        input_dep = func = lambda x: "good-slug"
        print_dep = func = lambda x: None
        #create the text file we'll use to replace stuff in
        with open("tests/nastifiles/mutation_run/test.txt", "w") as f:
            f.write("just replace_me please")
        #open the yaml file
        with open("tests/nastifiles/mutation_run/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_run", os, open, input_dep, print_dep)
        mutation.run()
        #open the file we replaced stuff in
        with open("tests/nastifiles/mutation_run/test.txt", "r") as f:
            file_contents = f.read()
        assert file_contents == "just good-slug please"
        #delete the file we replaced stuff in
        os.remove("tests/nastifiles/mutation_run/test.txt")

    def test_mutation_default_and_globals_simple(self):
        input_dep = func = lambda x: "bogus_slug"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/mutation_default_and_globals_simple/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation_config["globals"] = {config["globals"][0]["name"]: "Test App"}
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_default_and_globals_simple", os, open, input_dep, print_dep)
        default_value = mutation.render_default_template()
        assert default_value == "Test App"

    def test_mutation_default_and_globals_accept_default(self):
        input_dep = func = lambda x: "bogus_slug"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/mutation_default_and_globals_simple/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation_config["globals"] = {config["globals"][0]["name"]: "Test App"}
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_default_and_globals_simple", os, open, input_dep, print_dep, True)
        default_value = mutation.render_default_template()
        assert default_value == "Test App"

    def test_mutation_default_and_globals_complex(self):
        input_dep = func = lambda x: "bogus_slug"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/mutation_default_and_globals_complex/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation_config["globals"] = {config["globals"][0]["name"]: "Test App"}
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_default_and_globals_complex", os, open, input_dep, print_dep)
        default_value = mutation.render_default_template()
        assert default_value == "test_app"

    def test_mutation_default_exception(self):
        input_dep = func = lambda x: "bogus_slug"
        print_dep = func = lambda x: None
        #open the yaml file
        with open("tests/nastifiles/mutation_default_exception/nasti.yaml", "r") as f:
            config = yaml.safe_load(f)
        mutation_config = config["mutations"][0]
        mutation_config["globals"] = {config["globals"][0]["name"]: "Test App"}
        mutation = Mutation(mutation_config, "tests/nastifiles/mutation_default_exception", os, open, input_dep, print_dep)
        with self.assertRaises(exceptions.MutationDefaultTemplateInvalidException):
            mutation.render_default_template()


import os
import uuid
import nasti.exceptions as exceptions

# SourceHandlerResolver finds the correct source handler for the source input and returns it
# If no source is provided, it returns the help handler
class SourceHandlerResolver:
    HELP_KEY = "help"
    GIT_PREFIX = "git@"

    def __init__(self, source, help_text, os_dep=os, print_dep=print):
        self.os_dep = os_dep
        self.source = source
        self.print_dep = print_dep
        self.help_text = help_text

    def resolve(self):
        if not self.source or self.source.lower() == self.HELP_KEY:
            return HelpHandler(self.source, self.os_dep, self.print_dep, self.help_text)
        if self.source.startswith(self.GIT_PREFIX):
            return GitHandler(self.source, self.os_dep)
        return LocalDirectoryHandler(self.source, self.os_dep)

class GitHandler:
    source_dir = ""
    TMP_DIR = "/tmp/nasti/"
    def __init__(self, source, os_dep=os):
        self.source = source
        self.os_dep = os_dep

    def run(self):
        self.validate_git()
        self.__create_tmp_dir()
        self.clone_repo()

    def clean_up(self):
        self.os_dep.system('rm -rf ' + self.source_dir)
    
    def __command_exists(self, command):
        return self.os_dep.system("which " + command + " > /dev/null") == 0

    # I wouldn't normally make these methods public
    # but it aids in testing these methods
    def validate_git(self):
        if not self.__command_exists('git'):
            raise exceptions.GitHandlerGitMissingException("Error: git is not installed.")
    def clone_repo(self):
        if self.os_dep.system('git clone ' + self.source + ' ' + self.source_dir) != 0:
            raise exceptions.GitHandlerCloneException("Error: Unable to clone git repo.")

    def __create_tmp_dir(self):
        tmp_dir_path = self.TMP_DIR
        if not self.os_dep.path.isdir(tmp_dir_path):
            try:
                self.os_dep.mkdir(tmp_dir_path)
            except:
                raise exceptions.GitHandlerTmpDirCreationException("Error: Unable to create tmp directory.")
        random_dir = str(uuid.uuid4())
        self.source_dir = tmp_dir_path + random_dir
        try:
            self.os_dep.mkdir(self.source_dir)
        except:
            raise exceptions.GitHandlerTmpDirCreationException("Error: Unable to create tmp directory.")

class LocalDirectoryHandler:
    source_dir = ""
    def __init__(self, source, os_dep=os):
        self.source = source
        self.os_dep = os_dep
    
    def run(self):
        self.__validate_source_dir()
        self.source_dir = self.source

    def clean_up(self):
        pass

    def __validate_source_dir(self):
        if not self.os_dep.path.isdir(self.source):
            raise exceptions.LocalDirHandlerSourceNotDirException("Error: " + self.source + " is not a directory.")

class HelpHandler:
    source_dir = ""
    def __init__(self, source, os_dep=os, print_dep=print, help_text=""):
        self.source = source
        self.os_dep = os_dep
        self.print_dep = print_dep
        self.help_text = help_text

    def run(self):
        self.print_dep(self.help_text)

    def clean_up(self):
        pass




class MockPath:
    def __init__(self, opts):
        self.exists_ret = opts["exists"] if "exists" in opts else True
        self.join_ret = opts["join"] if "join" in opts else True
        self.isfile_ret = opts["isfile"] if "isfile" in opts else True
        self.isdir_ret = opts["isdir"] if "isdir" in opts else True
    def join(self, path1, path2):
        return self.join_ret
    def isfile(self, path):
        return self.isfile_ret
    def isdir(self, path):
        return self.isdir_ret
    def exists(self, path):
        return self.exists_ret

class MockOs:
    
    def __init__(self, opts={}):
        self.mkdir_ret = opts["mkdir"] if "mkdir" in opts else True
        self.system_ret = opts["system"] if "system" in opts else 0
        self.devnull_ret = opts["devnull"] if "devnull" in opts else True
        self.path = MockPath(opts)
        pass

    def mkdir(self, path):
        if not self.mkdir_ret:
            raise OSError("Error: Directory creation failed.")
        return self.mkdir_ret
    
    def system(self, command):
        return self.system_ret

    def devnull(self):
        return self.devnull_ret


def mock_print(text):
    pass

def mock_open(path, mode):
    pass

def mock_input(text):
    pass

def mock_failed_open(path, mode):
    raise OSError("Error: Unable to open file.")


import os
import re
from nasti.validation import Validation
import nasti.exceptions as exceptions
from jinja2 import Template, Environment
import jinja2.exceptions as jinja2_exceptions


class Mutation:
    prompt = ""
    help = False
    validation = False
    replace = ""
    files = []
    path = ""
    default = False
    globals = {}

    HELP_KEY = "help"
    VALIDATION_KEY = "validation"
    NAME_KEY = "name"
    PROMPT_KEY = "prompt"
    REPLACE_KEY = "replace"
    FILES_KEY = "files"
    DEFAULT_KEY = "default"
    GLOBALS_KEY = "globals"

    def __init__(self, mutation_config: dict, path, os_dep=os, open_dep=open, input_dep=input, print_dep=print, accept_defaults=False, silent_mode=False, silent_opts={}):
        # Dependency injection
        self.os_dep = os_dep
        self.open_dep = open_dep
        self.input_dep = input_dep
        self.print_dep = print_dep
        self.accept_defaults = accept_defaults
        self.silent_mode = silent_mode
        self.silent_opts = silent_opts
        # Required fields
        try:
            self.name        = mutation_config[self.NAME_KEY]
            self.prompt      = mutation_config[self.PROMPT_KEY]
            self.replace     = mutation_config[self.REPLACE_KEY]
            self.files       = mutation_config[self.FILES_KEY]
        except Exception as e:
            raise exceptions.MutationRequiredKeysMissingException(f"Error: Invalid mutation config: {mutation_config} required field missing missing {e}")
        # Optional fields
        if self.HELP_KEY in mutation_config:
            self.help        = mutation_config[self.HELP_KEY]
        if self.GLOBALS_KEY in mutation_config:
            self.globals     = mutation_config[self.GLOBALS_KEY]
        if self.DEFAULT_KEY in mutation_config:
            self.default     = mutation_config[self.DEFAULT_KEY]
        if self.VALIDATION_KEY in mutation_config:
            try:
                self.validation = Validation(mutation_config[self.VALIDATION_KEY])
            except Exception as e:
                raise e(f"Error: Invalid vaidation config in mutation: {mutation_config} {e}")
        self.path = path

    # Recursively find all files in the directory that contain the text to be replaced
    # and are not mentioned in the mutation
    # Returns a list of file paths relative to the working directory
    def find_unmentioned_files(self, root_dir):
        if not root_dir:
            root_dir = self.path
        unmentioned_files = []
        # list all files in the directory
        files_in_dir = self.os_dep.listdir(root_dir)
        # for each file in the directory
        for file_in_dir in files_in_dir:
            # skip the nastifile
            if file_in_dir == 'nasti.yaml':
                continue
            # if the file starts with a dot skip it
            if file_in_dir.startswith('.'):
                continue
            # The file path is relative to the working directory
            # so we need to construct fill file path
            file_full_path = root_dir + '/' + file_in_dir

            # if the file is a directory recurse
            if self.os_dep.path.isdir(file_full_path):
                unmentioned_files += self.find_unmentioned_files(file_full_path)
                continue
            
            # Determine if the file is in the mutation file list and if it is we'll skip it
            # this code is kind of hard to read so I'll explain it
            # We split the full file path by the working directory path
            # this gives us a list with elements: the working directory, the relative path to the file
            file_path_split = file_full_path.split(self.path)
            # remove the working directory from the list because we don't need it
            # as files in a mutation are relative to the working directory
            file_path_split.pop(0)
            # now we have a list with one element: the relative path to the file
            # we get the first element and remove the leading slash
            relative_file_path = file_path_split[0].lstrip('/')
            # the end result of all this is that a path like this:
            # /home/user/projects/nasti/nasti.yaml
            # with a working directory of:
            # /home/user/projects/nasti
            # becomes this:
            # nasti.yaml
            # which is the same format as the files in the mutation file list
            # so we can just check if the file is in the list and if it is we skip it
            if relative_file_path in self.files:
                continue

            # look into the file and see if the text to be replaced is there
            with self.open_dep(file_full_path, 'r', errors='replace') as f:
                if re.search(self.replace, f.read()):
                    # if the text to be replaced is there add it to the list
                    unmentioned_files.append(relative_file_path)
        return unmentioned_files

    # Validate the nastifile syntax
    # Raises exceptions if there are any problems
    def validate(self):
        # verify files isn't empty
        if len(self.files) == 0:
            raise exceptions.MutationEmptyFilesException(f"Error: mutation {self.name} does not contain any files.")
        # If there's a default template verify it renders
        if self.default:
            self.validate_default_template()
        for file in self.files:
            # verify file exists
            file_with_path = self.__get_file_full_path(file)
            if not self.os_dep.path.isfile(file_with_path):
                raise exceptions.MutationFileDoesNotExistException(f"Error: mutation: {self.name} file: {file} at: {file_with_path} does not exist.")
            # verify the file contains the text to be replaced
            with self.open_dep(file_with_path, 'r') as f:
                # search for at least one instance of the text to be replaced
                if not re.search(self.replace, f.read()):
                    raise exceptions.MutationFileDoesNotContainReplacementStringException(f"Error: mutation {self.name} file: {file} at: {file_with_path} does not contain {self.replace} ")

    def validate_default_template(self):
        default_value = ""
        if not self.default:
            return
        try:
            env = Environment()
            env.parse(self.default)
        except Exception as e:
            raise exceptions.MutationDefaultTemplateInvalidException(f"Error: Unable to render default template: {e}")
        return default_value

    def render_default_template(self):
        default_value = ""
        if not self.default:
            return
        try:
            template = Template(self.default)
            default_value = template.render(**self.globals)
        except Exception as e:
            raise exceptions.MutationDefaultTemplateInvalidException(f"Error: Unable to render default template: {e}")
        return default_value
        

    def run(self):
        # Handle silent mode
        if self.silent_mode:
            self.__run_silent()
            return
        # Handle accept defaults
        if self.accept_defaults and self.default:
            self.print_dep(f"[green]:heavy_check_mark:[/green] Using [green]{self.render_default_template()}[/green] for [blue]{self.name}[/blue] ")
            self.__replace_text_in_files(self.render_default_template())
            return
        # Prompt the user for input
        self.print_dep(f"{self.prompt}")
        if self.help:
            self.print_dep(self.help)
        # If there is a default value, print it
        if self.default:
            self.print_dep(f"[italic]Press enter for Default:[/italic] [blue]{self.render_default_template()}[/blue]")
        user_input = self.__get_user_input()
        try:
            self.__replace_text_in_files(user_input)
        except Exception as e:
            raise exceptions.MutationTextReplacementFailedException(f"Error: Unable to replace text in files: {e}")

    def __run_silent(self):
        # if there is no default and no silent opts, raise an exception
        if not self.default and not self.name in self.silent_opts:
            raise exceptions.MutationSilentModeException(f"Silent mode enabled but no default value or silent opts found for mutation: {self.name}")
        # If there is a default value, use it
        if self.default:
            new_value = self.render_default_template()
        # If there is no default value, use the value from silent opts
        if self.silent_opts and self.name in self.silent_opts:
            new_value = self.silent_opts[self.name]
        # Validate the new value
        if not self.__is_input_valid(new_value):
            raise exceptions.MutationSilentModeException(f"Invalid value found for mutation: {self.name}")
        try:
            self.__replace_text_in_files(new_value)
        except Exception as e:
            raise exceptions.MutationTextReplacementFailedException(f"Error: Unable to replace text in files: {e}")
        
    def __get_user_input(self):
        tries = 0
        max_tries = 3
        user_input = ""
        while True:
            user_input = self.input_dep("> ")
            tries += 1
            if tries > max_tries:
                raise exceptions.MutationTooManyInputTriesException(f"Error: Too many invalid inputs.")
            if self.default and user_input == "":
                user_input = self.render_default_template()
                break
            if self.__is_input_valid(user_input):
                break
            self.print_dep("[red]:x: Invalid input.[/red]")
        return user_input

    def __replace_text_in_files(self, user_input):
        # Replace the text in the files
        for file in self.files:
            file_with_path = self.__get_file_full_path(file)
            with self.open_dep(file_with_path, 'r') as f:
                file_text = f.read()
            file_text = file_text.replace(self.replace, user_input)
            with self.open_dep(file_with_path, 'w') as f:
                f.write(file_text)
    
    def __is_input_valid(self, user_input):
        # If there is no validation, the input is valid
        if not self.validation:
            return True
        # Validate the input by testing it against the validations
        return self.validation.validate(user_input)

    def __get_file_full_path(self, file):
        return self.path + '/' + file


class MutationRequiredKeysMissingException(Exception):
    pass

class MutationEmptyFilesException(Exception):
    pass

class MutationFileDoesNotExistException(Exception):
    pass

class MutationFileDoesNotContainReplacementStringException(Exception):
    pass

class MutationTooManyInputTriesException(Exception):
    pass

class MutationTextReplacementFailedException(Exception):
    pass

class MutationDefaultTemplateInvalidException(Exception):
    pass

class ValidationConfigMissingException(Exception):
    pass

class ValidationConfigInvalidException(Exception):
    pass

class ValidationUnknownKindException(Exception):
    pass

class GitHandlerGitMissingException(Exception):
    pass

class GitHandlerTmpDirCreationException(Exception):
    pass

class GitHandlerCloneException(Exception):
    pass

class LocalDirHandlerSourceNotDirException(Exception):
    pass

class NastiFileUnableToOpenFileException(Exception):
    pass

class NastiFileInvalidYamlException(Exception):
    pass

class NastiFileNoMutationsException(Exception):
    pass

class NastiFileUnknownKeysException(Exception):
    pass

class NastiFileGlobalNotFoundException(Exception):
    pass

class GlobalRequiredKeysMissingException(Exception):
    pass

class GlobalTooManyInputTriesException(Exception):
    pass

class GlobalSilentModeException(Exception):
    pass

class MutationSilentModeException(Exception):
    pass

class HooksScriptNotFound(Exception):
    pass
class HooksScriptExecutionFailed(Exception):
    pass
class HooksCleanupFailed(Exception):
    pass


from nasti.validation import Validation
import nasti.exceptions as exceptions

class Global:
    NAME_KEY = "name"
    PROMPT_KEY = "prompt"
    HELP_KEY = "help"
    VALIDATION_KEY = "validation"

    name = None
    prompt = None
    value = None
    help = None
    validation = None

    def __init__(self, opts: dict, input_dep, print_dep, silent_mode=False, silent_opts={}):
        # Dependency injection
        self.input_dep = input_dep
        self.print_dep = print_dep
        self. silent_mode = silent_mode
        self.silent_opts = silent_opts
        # Required fields
        try:
            self.name        = opts[self.NAME_KEY]
            self.prompt      = opts[self.PROMPT_KEY]
        except Exception as e:
            raise exceptions.GlobalRequiredKeysMissingException(f"Error: Invalid global config: {opts} required field missing: {e}")
        # Optional fields
        if self.HELP_KEY in opts:
            self.help        = opts[self.HELP_KEY]
        if self.VALIDATION_KEY in opts:
            try:
                self.validation = Validation(opts[self.VALIDATION_KEY])
            except Exception as e:
                raise e(f"Error: Invalid vaidation config in global: {opts} {e}")
    
    def validate(self):
        if self.validation:
            return self.validation.validate(self.value)
        return True

    def populate(self):
        if self.silent_mode:
            self.__populate_silent()
            return
        tries = 0
        max_tries = 3
        while True:
            self.print_dep(f"{self.prompt}")
            if self.help:
                self.print_dep(self.help)
            self.value = self.input_dep("> ")
            if self.validate():
                break
            tries += 1
            if tries >= max_tries:
                raise exceptions.GlobalTooManyInputTriesException(f"Error: Too many tries for global {self.name}")

    def __populate_silent(self):
        if self.name not in self.silent_opts:
            raise exceptions.GlobalSilentModeException(f"Error: Silent mode enabled but no value found for global {self.name}")
        self.value = self.silent_opts[self.name]
        if self.validate():
            return
        raise exceptions.GlobalSilentModeException(f"Error: Silent mode enabled but value {self.value} for global {self.name} failed validation")            

    def get_value(self):
        return self.value
    
    def get_name(self):
        return self.name



import unittest
from nasti.source_handlers import SourceHandlerResolver, GitHandler, LocalDirectoryHandler, HelpHandler
import tests.mocks as mocks
import nasti.exceptions as exceptions

class TestSourceHandlerResolver(unittest.TestCase):
    def test_resolve_dir_handler(self):
        mock_os = mocks.MockOs()
        resolver = SourceHandlerResolver("this_charming_man", "there is a light that never goes out", mock_os, mocks.mock_print)
        handler = resolver.resolve()
        assert isinstance(handler, LocalDirectoryHandler)

    def test_resolve_git_handler(self):
        mock_os = mocks.MockOs()
        resolver = SourceHandlerResolver("git@test.com/test/test.git", "help text", mock_os, mocks.mock_print)
        handler = resolver.resolve()
        assert isinstance(handler, GitHandler)
    
    def test_resolve_help_handler(self):
        mock_os = mocks.MockOs()
        resolver = SourceHandlerResolver(None, "We all go up in flames, going out in style", mock_os, mocks.mock_print)
        handler = resolver.resolve()
        assert isinstance(handler, HelpHandler)

class TestGitHandler(unittest.TestCase):
    def test_run(self):
        handler = GitHandler("git@test.com/test/test.git", mocks.MockOs())
        handler.run()
    
    def test_run_no_git(self):
        handler = GitHandler("git@test.com/test/test.git", mocks.MockOs({"system": 1}))
        with self.assertRaises(exceptions.GitHandlerGitMissingException):
            handler.validate_git()
    
    def test_run_create_tmp_dir_fails(self):
        handler = GitHandler(
            "git@test.ci/test/test.git", 
            mocks.MockOs({
                "mkdir": False, 
                "isdir": False
            })
        )
        with self.assertRaises(exceptions.GitHandlerTmpDirCreationException):
            handler.run()
    
    def test_run_clone_fails(self):
        handler = GitHandler(
            "git@test.com/test/test.git",
            mocks.MockOs({
                "system": 1
            })
        )
        with self.assertRaises(exceptions.GitHandlerCloneException):
            handler.clone_repo()

class TestLocalDirectoryHandler(unittest.TestCase):
    def test_run(self):
        handler = LocalDirectoryHandler("tests/mocks/test_dir", mocks.MockOs())
        handler.run()
    
    def test_run_no_dir(self):
        handler = LocalDirectoryHandler("tests/mocks/this_dir_does_not_exist", mocks.MockOs(
            {"isdir": False}
        ))
        with self.assertRaises(exceptions.LocalDirHandlerSourceNotDirException):
            handler.run()

# There is no test for HelpHandler because it is just a byzantine print statement.


import validators
import re
import nasti.exceptions as exceptions

class Validation:
    regex = False
    kind = False

    kinds = {
        "domain":       validators.domain,
        "email":        validators.email,
        "ip_address":   validators.ip_address,
        "slug":         validators.slug,
        "url":          validators.url,
        "uuid":         validators.uuid,
    }

    REGEX_KEY = "regex"
    KIND_KEY = "kind"

    def __init__(self, validation_config):
        try:
            self.__validate_config(validation_config)
        except Exception as e:
            raise e
        if self.REGEX_KEY in validation_config:
            self.regex = validation_config[self.REGEX_KEY]
        if self.KIND_KEY in validation_config:
            self.kind = validation_config[self.KIND_KEY]
    
    def validate(self, input_text):
        if self.regex:
            return self.__is_valid_regex(input_text)
        if self.kind:
            return self.__is_valid_kind(input_text)

    def __is_valid_regex(self, input_text):
        return bool(re.match(self.regex, input_text))
    
    def __is_valid_kind(self, input_text):
        return bool(self.kinds[self.kind](input_text))

    def __verify_known_kind(self, kind):
        if not kind in self.kinds:
            raise exceptions.ValidationUnknownKindException(f"Error: Unknown kind {kind}")

    def __validate_config(self, validation_config):
        if not "regex" in validation_config and not "kind" in validation_config: 
            raise exceptions.ValidationConfigMissingException(f"Error: Validation {validation_config} requires regex or kind")
        if "regex" in validation_config and "kind" in validation_config: 
            raise exceptions.ValidationConfigInvalidException(f"Error: Validation {validation_config} requires regex or kind, not both")
        if "kind" in validation_config:
            self.__verify_known_kind(validation_config["kind"])




import json

def load_dependencies_from_pipfile_lock():
    with open("Pipfile.lock", "r") as f:
        lock_content = json.load(f)
        default_packages = {pkg: details['version'] for pkg, details in lock_content["default"].items()}
        develop_packages = {pkg: details['version'] for pkg, details in lock_content["develop"].items()}
        return default_packages, develop_packages

default_requires, dev_requires = load_dependencies_from_pipfile_lock()

# Convert the dictionaries into list of strings with pinned versions
install_requires = [f"{pkg}{version}" for pkg, version in default_requires.items()]
dev_requires = [f"{pkg}{version}" for pkg, version in dev_requires.items()]


from setuptools import setup, find_packages

setup(
    name="NASTI",
    version="0.4.1",
    description="NASTI's A Strange Templating Implementation",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/adamrdrew/nasti",
    author="Adam Drew",
    author_email="adamrdrew@live.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
    },
    python_requires=">=3.10",
    entry_points={
        "console_scripts": [
            "nasti=nasti.cli:cli",
        ],
    },
)


import unittest
from nasti.nastifile import NastiFile, UnmentionedFilesResult, UnmentionedFilesResultItem
import tests.mocks as mocks
import os 
import nasti.exceptions as exceptions

class TestNastiFile(unittest.TestCase):
    def test_unable_to_open_file(self):
        nasti = NastiFile({
            "path": "tests/mocks/ddsfdsfdsfsdf",
            "os_dep": os,
            "open_dep": open,
        })
        with self.assertRaises(exceptions.NastiFileUnableToOpenFileException):
            nasti.load()

    def test_load_valid_yaml(self):
        nasti_file = NastiFile({
            "path": "tests/nastifiles/valid",
            "os_dep": os,
            "open_dep": open,
        })
        nasti_file.load()
        assert nasti_file.config["mutations"][0]["name"] == "example_mutation"

    def test_load_invalid_yaml(self):
        nasti_file = NastiFile({
            "path": "tests/nastifiles/bad_yaml",
            "os_dep": os,
            "open_dep": open,
        })
        with self.assertRaises(exceptions.NastiFileInvalidYamlException):
            nasti_file.load()
    
    def test_validate_no_mutations(self):
        nasti_file = NastiFile({
            "path": "tests/nastifiles/no_mutations",
            "os_dep": os,
            "open_dep": open,
        })
        nasti_file.load()
        with self.assertRaises(exceptions.NastiFileNoMutationsException):
            nasti_file.validate_mutations()
    
    def test_validate_mutations_unknown_keys(self):
        nasti_file = NastiFile({
            "path": "tests/nastifiles/mutation_unknown_keys",
            "os_dep": os,
            "open_dep": open,
        })
        nasti_file.load()
        with self.assertRaises(exceptions.NastiFileUnknownKeysException):
            nasti_file.validate_mutations()
    
    def test_unmentioned_files_found(self):
        nasti_file = NastiFile({
            "path": "tests/nastifiles/mutation_unmentioned_files",
            "os_dep": os,
            "open_dep": open,
        })
        unmentioned_files = nasti_file.find_unmentioned_files()
        # assert on array length
        assert isinstance(unmentioned_files, UnmentionedFilesResult)
        report = unmentioned_files.get_report()
        assert isinstance(report, str)
        assert len(report) != 0
        assert len(unmentioned_files.get_results()) == 1
        assert len(unmentioned_files.get_results()[0].files) == 2
        result = unmentioned_files.get_results()[0]
        assert isinstance(result, UnmentionedFilesResultItem)
        assert result.get_mutation().name == "example_mutation"
        assert "files/unmentioned" in result.get_files()
        assert "files/nested/unmentioned" in result.get_files()

    def test_unmentioned_files_not_found(self):
        nasti_file = NastiFile({
            "path": "tests/nastifiles/valid",
            "os_dep": os,
            "open_dep": open,
        })
        unmentioned_files = nasti_file.find_unmentioned_files()
        assert len(unmentioned_files.get_results()) == 0
        assert len(unmentioned_files.get_report()) == 0
    
    def test_globals(self):
        input_dep = func = lambda x: "test-global-value"
        print_dep = func = lambda x: None
        nasti_file = NastiFile({
            "path": "tests/nastifiles/nastifile_globals",
            "os_dep": os,
            "open_dep": open,
            "input_dep": input_dep,
            "print_dep": print_dep
        })
        nasti_file.load()
        nasti_file.run()
        global_obj = nasti_file.get_global("app_name")
        assert global_obj  == "test-global-value"
    
    def test_default(self):
        input_dep = func = lambda x: "input-from-user"
        print_dep = func = lambda x: None
        nasti_file = NastiFile({
            "path": "tests/nastifiles/nastifile_default",
            "os_dep": os,
            "open_dep": open,
            "input_dep": input_dep,
            "print_dep": print_dep
        })
        nasti_file.load()
        nasti_file.run()

    def test_run_greeting(self):
        input_dep = func = lambda x: "input_from_user"
        print_dep = func = lambda x: None
        nasti_file = NastiFile({
            "path": "tests/nastifiles/nastifile_greeting",
            "os_dep": os,
            "open_dep": open,
            "input_dep": input_dep,
            "print_dep": print_dep
        })
        nasti_file.load()
        nasti_file.run_greeting()
        assert nasti_file.config["greeting"] == "Welcome to the greeting!"
